/******************************\
*                              *
* EIGENVALUES AND EIGENVECTORS *
* ---------------------------- *
*                              *
*            PART II           *
*                              *
********************************
*                              *
*       DEFLATION METHOD       *
*        (General case)        *
*                              *
********************************
*                              *
*      Diochnos Dimitris       *
*          (stud1098)          *
*                              *
*      Saltampassis Nick       *
*          (stud1154)          *
*                              *
********************************
*                              *
*       3 - 5 June 2003        *
*                              *
\******************************/


/*
** ! Note: The arrays begin at index number 1 and NOT at index 0.
*/

#include "deflation.h"

/*********************************************************/

int		Deflation (PINAKASPTR matrix, const double tol, const int max_iterations, 
				   DIANYSMAPTR idiotimes, PINAKASPTR idiodianysmata)
{
	PINAKASPTR		L, L_Inverse, proswrinoulhs1, proswrinoulhs2;
	DIANYSMAPTR		temp_vector;
	int				size = matrix->size, i, j, p;
	PINAKASPTR *	Akolouthia = (PINAKASPTR *) malloc (sizeof (PINAKASPTR) * size);
	double			temp, max;

	temp_vector = Initialize_Dianysma (size);
	for (i = 1; i <= temp_vector->theseis; i++)
		Set_Dianysma (temp_vector, i, (double) 1.0);	// Initial estimation for max eigenvalue.

	Power_Method (matrix, tol, max_iterations, &temp, temp_vector);	// Calculate Max Eigenvalue.

	Set_Dianysma (idiotimes, 1, temp);	// Store MAX eigenvalue.
	for (i = 1; i <= size; i++) {		// Store MAX eigenvector.
		Get_Dianysma (temp_vector, i, &temp);
		Set_Pinaka (idiodianysmata, i, 1, temp);
	}

	/* Create empty the matrix sequence that will arise. */
	for (i = 0; i < size; i++)
		Akolouthia [i] = Initialize_Pinaka (size - i, size - i);

	/* Copy Initial matrix. */
	proswrinoulhs1 = Initialize_Pinaka (size, size);
	for (i = 1; i <= size; i++) {
		for (j = 1; j <= size; j++) {
			Get_Pinaka (matrix, i, j, &temp);
			Set_Pinaka (proswrinoulhs1, i, j, temp);
		}
	}

	p = 1;
	Get_Pinaka (idiodianysmata, 1, 1, &max);
	for (i = 2; i <= size; i++) {
		Get_Pinaka (idiodianysmata, i, 1, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}

	/******************************\
	* Simulate unitary matrix with *
	* line interchange.            *
	\******************************/

	if (p != 1) {
		Swap_Lines (proswrinoulhs1, 1, p);
		Swap_Columns_Pinaka (proswrinoulhs1, 1, p);
	}

	L = Initialize_Pinaka (size, size);
	L_Inverse = Initialize_Pinaka (size, size);

	Swap_Columns_Dianysma (temp_vector, 1, p);

	for (i = 1; i <= size; i++) {
		Set_Pinaka (L, i, i, (double) 1.0);
		Set_Pinaka (L_Inverse, i, i, (double) 1.0);
		if (i != 1) {
			Get_Dianysma (temp_vector, i, &temp);
			Set_Pinaka (L, i, 1, 0 - temp);
			Set_Pinaka (L_Inverse, i, 1, temp);
		}
	}

	proswrinoulhs2 = Initialize_Pinaka (size, size);
	
	Pollaplasiasmos_Pinakwn (L, proswrinoulhs1, proswrinoulhs2);
	Pollaplasiasmos_Pinakwn (proswrinoulhs2, L_Inverse, Akolouthia [0]);

	/*
	** Free up Memory.
	*/
	Destroy_Pinaka (proswrinoulhs1);
	Destroy_Pinaka (proswrinoulhs2);
	Destroy_Pinaka (L);
	Destroy_Pinaka (L_Inverse);
	Destroy_Dianysma (temp_vector);

	/*
	** Enter recursive call.
	*/
	Recursive_Deflation (Akolouthia, tol, max_iterations, idiotimes, idiodianysmata, 0);
	
	/*
	** Complete eigenvectors' co-ordinates.
	*/
	Inflate_Eigenvectors (Akolouthia, tol, max_iterations, size - 1, size, idiotimes, idiodianysmata);

	/*
	** Complete operation by destroying Akolouthia[N] Sequence.
	*/
	for (i = size - 1; i >= 0; i--)
		Destroy_Pinaka (Akolouthia [i]);
	free (Akolouthia);


	return OK;
}

/*********************************************************/
/*****************************************\
* * H synarthsh ayth ypologizei tis       *
* ypoloipes (N - 1) idiotimes tou arxikou *
* pinaka. Epipleon, den ypologizei plhrws *
* ta antistoixa idiodianysmata.           *
* * Sygkekrimena, an eimaste se ba8os:    *
* depth = k, ypologizei thn lambda_k      *
* idiotimh kai tis teleytaies (N - k - 1) *
* syntetagmenes pou antistoixoun sto      *
* antistoixo idiodianysma ths lambda_k.   *
* (O pinakas twn idiodianysmatwn einai    *
* Katw Trigwnikos otan oloklhrw8ei h      *
* diadikasia).                            *
* * Telos, h akolouthia pinakwn h opoia   *
* sxhmatizetai (Akolouthia [N]) 8a mas    *
* boh8hsei sto epomeno bhma na            *
* ypologisoume tis syntetagmenes poy      *
* leipoun stis opoies 8eseis toy pinaka   *
* twn idiodianysmatwn.                    *
\*****************************************/

int		Recursive_Deflation (PINAKASPTR * Akolouthia, const double tol, const int max_iterations, 
		DIANYSMAPTR idiotimes, PINAKASPTR idiodianysmata, const int depth)
{
	PINAKASPTR		L, L_Inverse, proswrinoulhs1, proswrinoulhs2;
	DIANYSMAPTR		temp_vector;
	int				size, i, j, p;
	double			temp, max;

	size = (int) Akolouthia [0]->size - depth - 1;

	temp_vector = Initialize_Dianysma (size);
	for (i = 1; i <= temp_vector->theseis; i++)
		Set_Dianysma (temp_vector, i, (double) 1.0);	// Initial estimation for max eigenvalue.

	/* Copy Initial matrix. */
	proswrinoulhs1 = Initialize_Pinaka (size, size);
	for (i = 2; i <= size + 1; i++) {
		for (j = 2; j <= size + 1; j++) {
			Get_Pinaka (Akolouthia [depth], i, j, &temp);
			Set_Pinaka (proswrinoulhs1, i - 1, j - 1, temp);
		}
	}

	Power_Method (proswrinoulhs1, tol, max_iterations, &temp, temp_vector);	// Calculate Max Eigenvalue.

	Set_Dianysma (idiotimes, depth + 2, temp);	// Store MAX eigenvalue.
	for (i = 1; i <= size; i++) {				// Store MAX eigenvector.
		Get_Dianysma (temp_vector, i, &temp);
		Set_Pinaka (idiodianysmata, i + depth + 1, depth + 2, temp);
	}

	if (depth == (int) Akolouthia [0]->size - 2) {
		Set_Pinaka (Akolouthia [Akolouthia [0]->size - 1], 1, 1, proswrinoulhs1->grammh [0]->sthlh [0]);

		/*
		** Free-Up Memory and complete recursion.
		*/
		Destroy_Pinaka (proswrinoulhs1);
		Destroy_Dianysma (temp_vector);

		return	OK;
	}

	p = depth + 2;
	Get_Pinaka (idiodianysmata, depth + 2, depth + 2, &max);
	for (i = depth + 2; i <= size + 1; i++) {
		Get_Pinaka (idiodianysmata, i, depth + 2, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}

	/******************************\
	* Simulate unitary matrix with *
	* line interchange.            *
	\******************************/

	if (p != depth + 2) {
		Swap_Lines (proswrinoulhs1, 1, p - depth - 1);
		Swap_Columns_Pinaka (proswrinoulhs1, 1, p - depth - 1);
	}

	L = Initialize_Pinaka (size, size);
	L_Inverse = Initialize_Pinaka (size, size);

	Swap_Columns_Dianysma (temp_vector, 1, p - depth - 1);

	for (i = 1; i <= size; i++) {
		Set_Pinaka (L, i, i, (double) 1.0);
		Set_Pinaka (L_Inverse, i, i, (double) 1.0);
		if (i != 1) {
			Get_Dianysma (temp_vector, i, &temp);
			Set_Pinaka (L, i, 1, 0 - temp);
			Set_Pinaka (L_Inverse, i, 1, temp);
		}
	}

	proswrinoulhs2 = Initialize_Pinaka (size, size);
	
	Pollaplasiasmos_Pinakwn (L, proswrinoulhs1, proswrinoulhs2);
	Pollaplasiasmos_Pinakwn (proswrinoulhs2, L_Inverse, Akolouthia [depth + 1]);

	/*
	** Free up Memory.
	*/
	Destroy_Pinaka (proswrinoulhs1);
	Destroy_Pinaka (proswrinoulhs2);
	Destroy_Pinaka (L);
	Destroy_Pinaka (L_Inverse);
	Destroy_Dianysma (temp_vector);


	/*
	** Go Deeper and deeper and ... deeeeeeeper!	:-)
	*/
	//depth++;

	Recursive_Deflation (Akolouthia, tol, max_iterations, idiotimes, idiodianysmata, depth + 1);

	return	OK;
}

/*********************************************************/
/*************************************\
* Oi sta8eres row kai column dhlwnoun *
* poio stoixeio apo tis syntetagmenes *
* twn idiodianysmatwn symplhrwnoume.  *
\*************************************/

int Inflate_Eigenvectors (PINAKASPTR * Akolouthia, const double tol, const int max_iterations, 
						  const int row, const int column, DIANYSMAPTR idiotimes, PINAKASPTR idiodianysmata)
{
	PINAKASPTR		L_Inverse;
	DIANYSMAPTR		temp_eigenvector, y, b, z;
	double			temp, lambda1, lambda2, max;
	int				i, j, p, size;

	size = (int) Akolouthia [0]->size;

	temp_eigenvector = Initialize_Dianysma (size - row + 1);
	z = Initialize_Dianysma (size - row + 1);
	y = Initialize_Dianysma (size - row);

	for (i = row; i <= size; i++) {
		Get_Pinaka (idiodianysmata, i, row, &temp);
		Set_Dianysma (temp_eigenvector, i - row + 1, temp);
		if (i > row) {
			Get_Pinaka (idiodianysmata, i, column, &temp);
			Set_Dianysma (z, i - row + 1, temp);
			Set_Dianysma (y, i - row, temp);
		}
	}

	b = Initialize_Dianysma (size - row);

	for (j = 2; j <= Akolouthia [row - 1]->size; j++) {
		Get_Pinaka (Akolouthia [row - 1], 1, j, &temp);
		Set_Dianysma (b, j - 1, temp);
	}

	Get_Dianysma (idiotimes, row, &lambda1);
	Get_Dianysma (idiotimes, column, &lambda2);
	Pollaplasiasmos_Dianysmatwn (b, y, &temp);

	Set_Dianysma (z, 1, (double) (temp / (lambda2 - lambda1)));

	/* Edw exei oloklhrw8ei o ypologismos toy z. */

	p = 1;
	Get_Dianysma (temp_eigenvector, 1, &max);
	for (i = 2; i <= temp_eigenvector->theseis; i++) {
		Get_Dianysma (temp_eigenvector, i, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}
	Swap_Columns_Dianysma (temp_eigenvector, 1, p);

	L_Inverse = Initialize_Pinaka (Akolouthia [row - 1]->size, Akolouthia [row - 1]->size);
	for (i = 1; i <= L_Inverse->size; i++) {
		Set_Pinaka (L_Inverse, i, i, (double) 1.0);
		if (i != 1) {
			Get_Dianysma (temp_eigenvector, i, &temp);
			Set_Pinaka (L_Inverse, i, 1, temp / max);
		}
	}
	if (p != 1)
		Swap_Lines (L_Inverse, 1, p);

	Pollaplasiasmos_Pinaka_Dianysmatos (L_Inverse, z, temp_eigenvector);

	p = 1;
	Get_Dianysma (temp_eigenvector, 1, &max);
	for (i = 2; i <= temp_eigenvector->theseis; i++) {
		Get_Dianysma (temp_eigenvector, i, &temp);
		if (fabs (temp) > fabs (max)) {
			p = i;
			max = temp;
		}
	}

	for (i = row; i <= size; i++) {
		Get_Dianysma (temp_eigenvector, i - row + 1, &temp);
		Set_Pinaka (idiodianysmata, i, column, temp / max);
	}

	Destroy_Pinaka (L_Inverse);
	Destroy_Dianysma (temp_eigenvector);
	Destroy_Dianysma (z);
	Destroy_Dianysma (y);
	Destroy_Dianysma (b);

	if ((column + 1) <= size)
		return	Inflate_Eigenvectors (Akolouthia, tol, max_iterations, row, column + 1, idiotimes, idiodianysmata);
	else if (row > 1)
		return	Inflate_Eigenvectors (Akolouthia, tol, max_iterations, row - 1, row, idiotimes, idiodianysmata);
	else
		return OK;
}

/*********************************************************/
