/******************************\
*                              *
* EIGENVALUES AND EIGENVECTORS *
* ---------------------------- *
*                              *
*            PART I            *
*                              *
********************************
*                              *
*         POWER METHOD         *
*                              *
********************************
*                              *
*      Diochnos Dimitris       *
*          (stud1098)          *
*                              *
*      Saltampassis Nick       *
*          (stud1154)          *
*                              *
********************************
*                              *
*       3 - 5 June 2003        *
*                              *
\******************************/


/*
** ! Note: The arrays begin at index number 1 and NOT at index 0.
*/


#include "power.h"

/*********************************************************/

int		Power_Method (PINAKASPTR matrix, const double tol, const int max_iterations, double * max_eigenvalue, DIANYSMAPTR eigenvector)
{
	int				size = (int) matrix->size, k = (int) 0, i, j;
	double			lambda0 = (double) 0.0, lambda1, result, temp, temp2, max;
	DIANYSMAPTR		y = Initialize_Dianysma (matrix->size), z = Initialize_Dianysma (matrix->size);

	/****************************************************\
	* ATTENTION: eigenvector comes as an initialized     *
	* ---------- argument of the function. It represents *
	* the initial estimation of the eigenvector.         *
	* Usually this estimation is the vector: <1,1,...,1> *
	\****************************************************/

	for (i = 1; i <= size; i++)	{
		/* Set initial estimation */
		Get_Dianysma (eigenvector, i, &result);
		Set_Dianysma (y, i, (double) result);
	}

	Get_Dianysma (y, 1, &max);
	for (i = 2; i <= size; i++) {
		Get_Dianysma (y, i, &result);
		if (fabs (result) > fabs (max))
			max = result;
	}

	for (i = 1; i <= size; i++) {
		Get_Dianysma (y, i, &result);
		Set_Dianysma (z, i, result / max);
	}

	while (k <= max_iterations) {
		for (i = 1; i <= size; i++) {
			result = (double) 0.0;
			for (j = 1; j <= size; j++) {
				Get_Pinaka (matrix, i, j, &temp);
				Get_Dianysma (z, j, &temp2);
				result += (double) temp * temp2;
			}
			Set_Dianysma (y, i, result);
		}

		Get_Dianysma (y, 1, &max);
		for (i = 2; i <= size; i++) {
			Get_Dianysma (y, i, &result);
			if (fabs (result) > fabs (max))
				max = result;
		}

		lambda1 = max;

		if (fabs (lambda1) < ZERO_TOLERANCE) {
			*max_eigenvalue = (double) 0.0;
			for (i = 1; i <= size; i++)
				Set_Dianysma (eigenvector, i, (double) 0.0);

			Destroy_Dianysma (y);		// De-Alloccate useless memory.
			Destroy_Dianysma (z);		// De-Alloccate useless memory.

			return	EIGENVALUE_ZERO;	// Indicate eigenvalue was Zero. Try another initial vector y.
		}

		for (i = 1; i <= size; i++) {
			Get_Dianysma (y, i, &result);
			Set_Dianysma (z, i, result / max);
		}

		if (fabs (lambda0 - lambda1) < tol) {
			*max_eigenvalue = (double) lambda1;
			for (i = 1; i <= size; i++)	{
				Get_Dianysma (z, i, &temp);
				Set_Dianysma (eigenvector, i, (double) temp);
			}

			Destroy_Dianysma (y);		// De-Alloccate useless memory.
			Destroy_Dianysma (z);		// De-Alloccate useless memory.

			return	OK;		// Indicate maximum absolute eigenvalue was found!
		}

		lambda0 = (double) lambda1;
		k++;
	}

	Destroy_Dianysma (y);
	Destroy_Dianysma (z);
	max_eigenvalue = (double *) NULL;
	for (i = 1; i <= size; i++)
		Set_Dianysma (eigenvector, i, (double) 0.0);

	return	POWER_DID_NOT_CONVERGE;
}
